# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2021/11/4 17:39
describe: A股强势股票传感器
"""
import os
import os.path
import traceback
import inspect
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from datetime import timedelta, datetime
from collections import Counter
from tqdm import tqdm
from typing import Callable
from czsc.objects import Event
from czsc.utils import io
from czsc.data.ts_cache import TsDataCache, Freq
from czsc.sensors.utils import get_index_beta, generate_signals, max_draw_down, turn_over_rate
from czsc.utils import WordWriter


plt.style.use('ggplot')
plt.rcParams['font.sans-serif'] = ['SimHei']
plt.rcParams['axes.unicode_minus'] = False


def selected_filter_by_index(dc: TsDataCache, dfg: pd.DataFrame, index_code=None):
    """使用指数成分过滤

    :param dc: 数据缓存对象
    :param dfg: 单个交易日的强势股选股结果
    :param index_code: 指数代码
    :return: 过滤后的选股结果
    """
    if not index_code or dfg.empty:
        return dfg

    assert dfg['trade_date'].nunique() == 1
    trade_date = dfg['trade_date'].max()

    index_members = dc.index_weight(index_code, trade_date)
    ts_codes = list(index_members['con_code'].unique())
    return dfg[dfg.ts_code.isin(ts_codes)]


def selected_filter_by_concepts(dc, dfg, top_n=20, min_n=3, method='v1'):
    """使用板块效应过滤

    :param dc: 数据缓存对象
    :param dfg: 单个交易日的强势股选股结果
    :param top_n: 选取前 n 个密集概念
    :param min_n: 单股票至少要有 n 个概念在 top_n 中
    :param method: 打分计算方法
        v1  直接取板块中的强势股数量作为分数
        v2  板块内强势股数 / 板块内股数
    :return: 过滤后的选股结果
    """
    if dfg.empty or not top_n or not min_n:
        return dfg, []

    ths_members = dc.get_all_ths_members(exchange="A", type_="N")
    ths_members = ths_members[~ths_members['概念名称'].isin([
        'MSCI概念', '沪股通', '深股通', '融资融券', '上证180成份股', '央企国资改革',
        '标普道琼斯A股', '中证500成份股', '上证380成份股', '沪深300样本股',
    ])]

    ths_concepts = ths_members[ths_members.code.isin(dfg.ts_code)]
    if method == 'v1':
        key_concepts = [k for k, v in Counter(ths_concepts['概念名称'].to_list()).most_common(top_n)]
    elif method == 'v2':
        all_count = Counter(ths_members['概念名称'].to_list())
        sel_count = Counter(ths_concepts['概念名称'].to_list())
        df_scores = pd.DataFrame([{"concept": k, 'score': sel_count[k] / all_count[k]}
                                  for k in sel_count.keys()])
        key_concepts = df_scores.sort_values('score', ascending=False).head(top_n)['concept'].to_list()
    else:
        raise ValueError(f"method value error")

    sel = ths_concepts[ths_concepts['概念名称'].isin(key_concepts)]
    ts_codes = [k for k, v in Counter(sel.code).most_common() if v >= min_n]
    dfg = dfg[dfg.ts_code.isin(ts_codes)]
    dfg.loc[:, '概念板块'] = dfg.ts_code.apply(lambda x: ths_concepts[ths_concepts.code == x]['概念名称'].to_list())
    dfg.loc[:, '概念数量'] = dfg['概念板块'].apply(len)

    return dfg, key_concepts


def selected_filter_by_market_value(dfg, min_total_mv=None):
    """使用总市值过滤

    :param dfg: 单个交易日的强势股选股结果
    :param min_total_mv: 最小总市值，单位为万元，1e6万元 = 100亿
    :return: 过滤后的选股结果
    """
    if dfg.empty or not min_total_mv:
        return dfg

    return dfg[dfg['total_mv'] >= min_total_mv]


def selected_filter_by_rps(dfg, n=21, v_range=(0.2, 0.8), max_count=-1):
    """使用b20b过滤，b20b 表示前20个交易日的涨跌幅

    :param dfg: 单个交易日的强势股选股结果
    :param n: RPS的计算区间
    :param v_range: RPS值按从大到小排序后的可选区间
        默认为 0.2 ~ 0.8，表示可选区间为排序位置在 20% ~ 80% 区间的股票
    :param max_count: 最多保留结果数量
    :return: 过滤后的选股结果
    """
    if dfg.empty or (not max_count) or len(dfg) < max_count:
        return dfg
    rps_col = f"b{n}b"
    # dfg = dfg.sort_values(rps_col, ascending=True)
    # dfg = dfg.reset_index(drop=True)
    # dfg = dfg.iloc[int(len(dfg) * v_range[0]): int(len(dfg) * v_range[1])]
    # return dfg.tail(max_count)

    split = v_range[1]
    dfg = dfg.sort_values(rps_col, ascending=True)
    head_i = int((len(dfg) - max_count) * split) + 1
    tail_i = len(dfg) - int((len(dfg) - max_count) * (1 - split))

    return dfg.iloc[head_i: tail_i]


def create_next_positions(dc: TsDataCache, dfg: pd.DataFrame):
    """构建某天选股结果对应的下一交易日持仓明细

    :param dc: 数据缓存对象
    :param dfg: 单个交易日的强势股选股结果
    :return: 下一交易日持仓明细
    """
    if dfg.empty:
        return dfg

    trade_cal = dc.trade_cal()
    trade_cal = trade_cal[trade_cal.is_open == 1]
    trade_dates = trade_cal.cal_date.to_list()
    trade_date = dfg['trade_date'].iloc[0]

    hold = dfg.copy()
    hold['成分日期'] = trade_dates[trade_dates.index(trade_date.strftime("%Y%m%d")) + 1]
    hold['持仓权重'] = 0.98 / len(dfg)
    hold.rename({'ts_code': "证券代码", "close": "交易价格"}, inplace=True, axis=1)
    hold = hold[['证券代码', '持仓权重', '交易价格', '成分日期']]
    hold['成分日期'] = pd.to_datetime(hold['成分日期']).apply(lambda x: x.strftime("%Y/%m/%d"))
    return hold


def plot_alpha_v1(beta_name, df_alpha, file_png) -> None:
    """用三个并列线图来绘制 alpha 信息

    :param beta_name: 基准指数名称
    :param df_alpha: 包含 ['trade_date', 'beta', 'selector']
          trade_date     beta    selector
        0 2018-01-02  88.4782   93.471190
        1 2018-01-03  45.8368   41.008785
        2 2018-01-04  -0.4383 -132.660895
        3 2018-01-05  45.0786  120.726060
        4 2018-01-08  -0.6757  -17.231665
    :param file_png: 图片保存文件名
    :return: None
    """
    plt.close()
    fig, axes = plt.subplots(nrows=3, ncols=1, figsize=(9, 5*3))
    df_alpha['beta_curve'] = df_alpha['beta'].cumsum()
    df_alpha['selector_curve'] = df_alpha['selector'].cumsum()
    df_alpha['alpha_curve'] = df_alpha['selector_curve'] - df_alpha['beta_curve']

    df_alpha.rename({'trade_date': 'date', 'beta_curve': f"beta_curve:{beta_name}"}, inplace=True, axis=1)
    for i, col in enumerate(['alpha_curve', 'selector_curve', f"beta_curve:{beta_name}"], 0):
        ax = axes[i]
        sns.lineplot(x='date', y=col, data=df_alpha, ax=ax)
        ax.text(x=df_alpha['date'].iloc[0], y=int(df_alpha[col].mean()),
                s=f"{col}：{int(df_alpha[col].iloc[-1])}", fontsize=12)
        ax.set_title(f"{col}", loc='center')
        ax.set_xlabel("")
    plt.savefig(file_png, bbox_inches='tight', dpi=100)
    plt.close()


def plot_alpha_v2(beta_name, df_alpha, file_png) -> None:
    """用线图来绘制 alpha 信息

    :param beta_name: 基准指数名称
    :param df_alpha: 包含 ['trade_date', 'beta', 'selector']
          trade_date     beta    selector
        0 2018-01-02  88.4782   93.471190
        1 2018-01-03  45.8368   41.008785
        2 2018-01-04  -0.4383 -132.660895
        3 2018-01-05  45.0786  120.726060
        4 2018-01-08  -0.6757  -17.231665
    :param file_png: 图片保存文件名
    :return: None
    """
    df_alpha['beta_curve'] = df_alpha['beta'].cumsum()
    df_alpha['selector_curve'] = df_alpha['selector'].cumsum()
    df_alpha['alpha_curve'] = df_alpha['selector_curve'] - df_alpha['beta_curve']
    df_alpha.rename({'trade_date': 'date', 'beta_curve': f"beta_curve:{beta_name}"}, inplace=True, axis=1)

    plt.close()
    plt.figure(figsize=(9, 5))
    sns.lineplot(x='date', y='alpha_curve', data=df_alpha)
    sns.lineplot(x='date', y='selector_curve', data=df_alpha)
    sns.lineplot(x='date', y=f"beta_curve:{beta_name}", data=df_alpha)
    plt.legend(labels=['超额', '选股', f"基准{beta_name}"])
    plt.savefig(file_png, bbox_inches='tight', dpi=100)


def plot_alpha_v3(beta_name, df_alpha, file_png) -> None:
    """用类似MACD图来绘制 alpha 信息

    :param beta_name: 基准指数名称
    :param df_alpha: 包含 ['trade_date', 'beta', 'selector']
          trade_date     beta    selector
        0 2018-01-02  88.4782   93.471190
        1 2018-01-03  45.8368   41.008785
        2 2018-01-04  -0.4383 -132.660895
        3 2018-01-05  45.0786  120.726060
        4 2018-01-08  -0.6757  -17.231665
    :param file_png: 图片保存文件名
    :return: None
    """
    df_alpha['beta_curve'] = df_alpha['beta'].cumsum()
    df_alpha['selector_curve'] = df_alpha['selector'].cumsum()
    df_alpha['alpha'] = df_alpha['selector'] - df_alpha['beta']
    df_alpha['alpha_curve'] = df_alpha['selector_curve'] - df_alpha['beta_curve']
    df_alpha.rename({'trade_date': 'date', 'beta_curve': f"beta_curve:{beta_name}"}, inplace=True, axis=1)

    plt.close()
    plt.figure(figsize=(9, 5))
    x = df_alpha['date']
    plt.bar(x, height=df_alpha['alpha'], width=0.01, color='blue', label='alpha')
    plt.plot(x, df_alpha['alpha_curve'], label='alpha_curve')
    plt.plot(x, df_alpha['selector_curve'], label='selector_curve')
    plt.plot(x, df_alpha[f"beta_curve:{beta_name}"], label=f"beta_curve:{beta_name}")
    plt.legend()
    plt.savefig(file_png, bbox_inches='tight', dpi=100)


class StocksDaySensor:
    """以日线为基础周期的强势股票感应器

    输入：市场个股全部行情、概念板块成分信息
    输出：强势个股列表以及概念板块分布
    """

    def __init__(self,
                 experiment_path: str,
                 sdt: str,
                 edt: str,
                 dc: TsDataCache,
                 strategy: Callable,
                 signals_n: int = 0,
                 ):

        self.name = self.__class__.__name__
        self.version = "V20220404"

        self.experiment_path = experiment_path
        self.results_path = os.path.join(experiment_path, f"{strategy()[1]().name}_{sdt}_{edt}")
        self.signals_path = os.path.join(experiment_path, 'signals')
        os.makedirs(self.experiment_path, exist_ok=True)
        os.makedirs(self.results_path, exist_ok=True)
        os.makedirs(self.signals_path, exist_ok=True)

        self.sdt = sdt
        self.edt = edt
        self.verbose = os.environ.get('verbose', False)

        self.strategy = strategy
        self.signals_n = signals_n
        self.get_signals, self.get_event = strategy()
        self.event: Event = self.get_event()
        self.base_freq = Freq.D.value
        self.freqs = [Freq.W.value, Freq.M.value]

        self.file_docx = os.path.join(self.results_path, f'{self.event.name}_{sdt}_{edt}.docx')
        writer = WordWriter(self.file_docx)
        if not os.path.exists(self.file_docx):
            writer.add_title("股票选股强度验证")
            writer.add_page_break()
            writer.add_heading(f"{datetime.now().strftime('%Y-%m-%d %H:%M')} {self.event.name}", level=1)

            writer.add_heading("参数配置", level=2)
            writer.add_paragraph(f"测试方法描述：{self.event.name}")
            writer.add_paragraph(f"测试起止日期：{sdt} ~ {edt}")
            writer.add_paragraph(f"信号计算函数：\n{inspect.getsource(self.get_signals)}")
            writer.add_paragraph(f"事件具体描述：\n{inspect.getsource(self.get_event)}")
            writer.save()

            with open(os.path.join(self.results_path, f"{strategy.__name__}.txt"), mode='w') as f:
                f.write(inspect.getsource(strategy))

        self.writer = writer
        self.dc = dc
        self.betas = ['000905.SH', '000300.SH', '399006.SZ']
        get_index_beta(dc, sdt, edt, freq='D', indices=self.betas,
                       file_xlsx=os.path.join(self.results_path, 'betas.xlsx'))
        file_dfm = os.path.join(self.results_path, f'df_event_matched_{sdt}_{edt}.pkl')
        file_dfb = os.path.join(self.experiment_path, f'df_all_bars_{sdt}_{edt}.pkl')
        if os.path.exists(file_dfm):
            self.dfm = io.read_pkl(file_dfm)
            self.dfb = io.read_pkl(file_dfb)
        else:
            self.dfm, self.dfb = self.get_stock_strong_days()
            io.save_pkl(self.dfm, file_dfm)
            io.save_pkl(self.dfb, file_dfb)
        self.nb_cols = [x for x in self.dfb.columns if x[0] == 'n' and x[-1] == 'b']

    def get_share_strong_days(self, ts_code: str, name: str):
        """获取单个标的全部强势信号日期"""
        dc = self.dc
        event = self.event
        sdt = self.sdt
        edt = self.edt
        file_signals = os.path.join(self.signals_path, f"{ts_code}.pkl")

        if os.path.exists(file_signals):
            signals, n_bars = io.read_pkl(file_signals)
            if self.verbose:
                print(f"get_share_strong_days: load signals from {file_signals}")
        else:
            start_date = pd.to_datetime(self.sdt) - timedelta(days=3000)
            bars = dc.pro_bar(ts_code=ts_code, start_date=start_date, end_date=edt, freq='D', asset="E", raw_bar=True)
            n_bars = dc.pro_bar(ts_code=ts_code, start_date=sdt, end_date=edt, freq='D', asset="E", raw_bar=False)
            signals = generate_signals(bars, sdt, self.base_freq, self.freqs, self.get_signals,
                                       signals_n=self.signals_n)
            io.save_pkl([signals, n_bars], file_signals)

        nb_dicts = {row['trade_date'].strftime("%Y%m%d"): row for row in n_bars.to_dict("records")}
        event_matched = []
        for s in signals:
            m, f = event.is_match(s)
            if m:
                nb_info = nb_dicts.get(s['dt'].strftime("%Y%m%d"), None)
                r = {'name': name, 'event_match': True, 'factor_match': f}
                if nb_info:
                    r.update(nb_info)
                    event_matched.append(r)

        dfs = pd.DataFrame(event_matched)
        if event_matched:
            df_ = dc.daily_basic(ts_code, sdt, dc.edt)
            df_['trade_date'] = pd.to_datetime(df_['trade_date'])
            dfs = dfs.merge(df_[['trade_date', 'total_mv']], on='trade_date', how='left')
            dfs = dfs[pd.to_datetime(sdt) <= dfs['trade_date']]
            dfs = dfs[dfs['trade_date'] <= pd.to_datetime(edt)]
        print(f"{ts_code} - {name}: {len(dfs)}")
        return dfs, n_bars

    def get_stock_strong_days(self):
        """获取全部股票的强势日期"""
        stocks = self.dc.stock_basic()

        all_matched = []
        all_bars = []
        for row in tqdm(stocks.to_dict('records'), desc="get_stock_strong_days"):
            ts_code = row['ts_code']
            name = row['name']
            try:
                dfs, n_bars = self.get_share_strong_days(ts_code, name)
                all_matched.append(dfs)
                all_bars.append(n_bars)
            except:
                print(f"get_share_strong_days error: {ts_code}, {name}")
                traceback.print_exc()

        dfm = pd.concat(all_matched, ignore_index=True)
        dfb = pd.concat(all_bars, ignore_index=True)
        return dfm, dfb

    def get_selected(self,
                     index_code=None,
                     fc_top_n=None,
                     fc_min_n=None,
                     min_total_mv=None,
                     max_count=None):
        """验证传感器在一组过滤参数下的表现

        :param index_code: 指数成分过滤
        :param fc_top_n: 板块效应过滤参数1
        :param fc_min_n: 板块效应过滤参数2
        :param min_total_mv: 市值效应过滤参数
        :param max_count: 控制最大选出数量
        :return:
        """
        dc = self.dc
        df = self.dfm

        selected_dfg = {}
        for trade_date, dfg in df.groupby('trade_date'):
            try:
                if dfg.empty:
                    print(f"{trade_date} 选股结果为空")
                    continue
                dfg, key_concepts = selected_filter_by_concepts(dc, dfg, top_n=fc_top_n, min_n=fc_min_n)
                dfg = selected_filter_by_market_value(dfg, min_total_mv)
                dfg = selected_filter_by_index(dc, dfg, index_code)
                dfg = selected_filter_by_rps(dfg, n=21, v_range=(0.1, 0.8), max_count=max_count)
                selected_dfg[trade_date] = {'dfg': dfg, 'key_concepts': key_concepts}
            except:
                traceback.print_exc()

        return selected_dfg

    def get_agg_selected(self, selected_dfg, window_size: int = 1):
        """滑动窗口聚合选股结果

        :param selected_dfg: get_selected 输出的结果
        :param window_size: 聚合窗口大小
        :return:
        """
        dc = self.dc
        dfb = self.dfb
        assert window_size > 0
        if window_size > 1:
            trade_cal = dc.trade_cal()
            trade_cal = trade_cal[trade_cal['is_open'] == 1].cal_date.to_list()
            trade_cal = [pd.to_datetime(x) for x in trade_cal]

            selected_agg = {}
            for td, dfg in tqdm(selected_dfg.items(), desc='agg_by_window'):
                i = trade_cal.index(td)
                windows = trade_cal[i-window_size+1: i+1]
                res = []
                for td_ in windows:
                    dfg = selected_dfg.get(td_, None)
                    if dfg:
                        df = dfg['dfg']
                        df['trade_date'] = td
                        res.append(dfg['dfg'])
                dfd = pd.concat(res, ignore_index=True)
                dfd_cols = ['name', 'event_match', 'factor_match', 'ts_code', 'trade_date',
                            'total_mv', '概念板块', '概念数量']
                dfd = dfd.drop_duplicates('ts_code', ignore_index=True)[dfd_cols]
                dfd = dfd.merge(dfb, on=['ts_code', 'trade_date'], how='left')
                selected_agg[td] = dfd
        else:
            selected_agg = {dt: x['dfg'] for dt, x in selected_dfg.items()}

        return selected_agg

    def validate_performance(self,
                             index_code=None,
                             fc_top_n=None,
                             fc_min_n=None,
                             min_total_mv=None,
                             max_count=None,
                             window_size=1,
                             save: bool = False,
                             ):
        """验证传感器在一组过滤参数下的表现

        :param index_code: 指数成分过滤
        :param fc_top_n: 板块效应过滤参数1
        :param fc_min_n: 板块效应过滤参数2
        :param min_total_mv: 市值效应过滤参数
        :param max_count: b20b过滤参数，控制最大选出数量
        :param window_size: 按 window_size 聚合多天的结果到一天
        :param save: 是否保存结果到本地
        :return:
        """
        dc = self.dc
        sdt = self.sdt
        edt = self.edt
        selected_dfg = self.get_selected(index_code, fc_top_n, fc_min_n, min_total_mv, max_count)
        selected_agg = self.get_agg_selected(selected_dfg, window_size)

        # 分析
        performances = []
        for td, df in selected_agg.items():
            p = {'trade_date': td, "number": len(df)}
            if df.empty:
                p.update({x: 0 for x in self.nb_cols})
            else:
                p.update(df[self.nb_cols].mean().to_dict())
            performances.append(p)

        df_p = pd.DataFrame(performances)
        df_detail = pd.concat([v for k, v in selected_agg.items()])
        df_holds = pd.concat([create_next_positions(dc, v) for k, v in selected_agg.items()], ignore_index=True)
        df_turns, tor = turn_over_rate(df_holds)

        beta = get_index_beta(dc, sdt, edt, freq='D', file_xlsx=None, indices=self.betas)
        df_n1b = pd.DataFrame()
        for name, df_ in beta.items():
            if name in self.betas:
                df_n1b['trade_date'] = pd.to_datetime(df_.trade_date.to_list())
                df_n1b[name] = df_.n1b.to_list()

        df_ = df_p[['trade_date', 'number', 'n1b']]
        df_.rename({'n1b': 'selector'}, axis=1, inplace=True)
        df_.reset_index(drop=True, inplace=True)
        df_n1b = df_n1b.merge(df_, on='trade_date', how='left')
        df_n1b.fillna(0, inplace=True)

        mdd_info = []
        for col in self.betas + ['selector']:
            df_n1b[f"{col}_curve"] = df_n1b[col].cumsum()
            df_n1b[f"{col}_curve"] += 10000
            start_i, end_i, mdd = max_draw_down(df_n1b[col].to_list())
            start_dt = df_n1b.iloc[start_i]['trade_date']
            end_dt = df_n1b.iloc[end_i]['trade_date']
            row = {
                '标的': col,
                "开始日期": sdt,
                "结束日期": edt,
                "累计收益": round(df_n1b[col].sum(), 4),
                "单笔收益": round(df_n1b[col].mean(), 4),
                "交易胜率": round(len(df_n1b[df_n1b[col] >= 0]) / len(df_n1b), 4),
                "最大回撤": mdd,
                "回撤开始日期": start_dt,
                "回撤结束日期": end_dt,
            }
            mdd_info.append(row)

        df_mdd = pd.DataFrame(mdd_info)

        if save:
            output_path = os.path.join(self.results_path,
                                       f"{index_code}_{fc_top_n}_{fc_min_n}_{min_total_mv}_{max_count}_{window_size}")
            os.makedirs(output_path, exist_ok=True)
            f = pd.ExcelWriter(os.path.join(output_path, 'performances.xlsx'))
            df_turns.to_excel(f, sheet_name="每日换手", index=False)
            df_p.to_excel(f, sheet_name="每日统计", index=False)
            df_n1b.to_excel(f, sheet_name="资金曲线", index=False)
            df_mdd.to_excel(f, sheet_name="性能对比", index=False)
            f.close()

            df_detail.to_csv(os.path.join(output_path, 'selected_details.csv'), index=False)
            df_holds.to_csv(os.path.join(output_path, 'holds.csv'), index=False)

        return df_detail, df_p, df_holds, df_n1b, df_mdd, tor

    def get_latest_selected(self,
                            index_code=None,
                            fc_top_n=None,
                            fc_min_n=None,
                            min_total_mv=None,
                            max_count=None,
                            window_size=1,
                            ):
        """获取下一个交易日的强势股列表

        :param index_code:
        :param fc_top_n:
        :param fc_min_n:
        :param min_total_mv:
        :param max_count:
        :param window_size:
        :return:
        """
        output = self.validate_performance(index_code, fc_top_n, fc_min_n, min_total_mv, max_count, window_size)
        df_detail, df_p, df_holds, df_n1b, df_mdd, tor = output
        df_holds['成分日期'] = pd.to_datetime(df_holds['成分日期'])
        max_date = df_holds['成分日期'].max()
        df_latest = df_holds[df_holds['成分日期'] == max_date]
        return df_latest

    def grip_search(self, grid_params=None):
        """网格搜索超参数

        :param grid_params: 网格参数
        :return:
        """
        if not grid_params:
            grid_params = {
                "fc_top_n": list(range(10, 160, 10)),
                "fc_min_n": list(range(1, 7, 1)),
                "min_total_mv": [5e5, 10e5, 20e5],
                "max_count": list(range(100, 300, 100)),
            }

        filter_params = [{
            'fc_top_n': None,
            'fc_min_n': None,
            'min_total_mv': None,
            'max_count': None,
        }]
        window_size = 1

        for i in grid_params['fc_top_n']:
            for j in grid_params['fc_min_n']:
                for k in grid_params['min_total_mv']:
                    for m in grid_params['max_count']:
                        filter_params.append({
                            'fc_top_n': i,
                            'fc_min_n': j,
                            'min_total_mv': k,
                            'max_count': m,
                        })

        results = []
        for conf in tqdm(filter_params):
            index_code = None
            fc_top_n = conf['fc_top_n']
            fc_min_n = conf['fc_min_n']
            min_total_mv = conf['min_total_mv']
            max_count = conf['max_count']
            try:
                output = self.validate_performance(index_code, fc_top_n, fc_min_n,
                                                   min_total_mv, max_count, window_size, save=False)
                df_detail, df_merged, df_holds, df_n1b, df_mdd, tor = output
                res = dict(conf)
                nb_keys = [x for x in df_merged.columns if x[0] == 'n' and x[-1] == 'b']
                for n in nb_keys:
                    res[n] = round(df_merged[n].mean(), 2)

                res['max_count'] = df_merged['number'].max()
                res['min_count'] = df_merged['number'].min()
                res['avg_count'] = int(df_merged['number'].mean())
                res['turn_over'] = tor

                results.append(res)

                df = pd.DataFrame(results)
                file_grid_res = os.path.join(self.results_path, f'{self.event.name}_grid_search_results.xlsx')
                df.to_excel(file_grid_res, index=False)
                print(f"grid search results saved into {file_grid_res}")
            except:
                traceback.print_exc()

    def write_validate_report(self, title, conf):
        """编写试验报告到 Word 文件

        :param title: 试验组别的标题
        :param conf: 验证的参数
        :return: None
        """
        results_path = self.results_path
        writer = self.writer
        index_code = conf.get('index_code', None)
        fc_top_n = conf.get('fc_top_n', None)
        fc_min_n = conf.get('fc_min_n', None)
        min_total_mv = conf.get('min_total_mv', None)
        max_count = conf.get('max_count', None)
        window_size = conf.get('window_size', 1)

        output = self.validate_performance(index_code, fc_top_n, fc_min_n,
                                           min_total_mv, max_count, window_size, save=True)
        df_detail, df_merged, df_holds, df_n1b, df_mdd, tor = output

        writer.add_page_break()
        writer.add_heading(title, level=1)
        writer.add_df_table(pd.DataFrame({"参数名称": list(conf.keys()), '参数值': list(conf.values())}))
        writer.add_paragraph('结果对比：')
        writer.add_df_table(df_mdd)
        msg = f"组合换手率：{tor}；最大持仓数：{df_merged['number'].max()}；" \
              f"平均持仓数：{int(df_merged['number'].mean())}；最小持仓数：{df_merged['number'].min()}"
        writer.add_paragraph(msg)
        writer.save()

        # 绘制曲线对比图
        for beta_name in self.betas:
            df_alpha = df_n1b[['trade_date', beta_name, 'selector']]
            df_alpha.rename({beta_name: 'beta'}, axis=1, inplace=True)
            file_png = os.path.join(results_path, f"{beta_name}_alpha.png")
            plot_alpha_v2(beta_name, df_alpha, file_png)

            writer.add_heading(f"与{beta_name}进行对比", level=2)
            writer.add_picture(file_png, width=15, height=9)
            writer.save()
            os.remove(file_png)




# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2022/2/7 19:05
"""
from datetime import datetime
from collections import OrderedDict
from czsc import CZSC, Signal


def get_s_raw_bar_end(c: CZSC, k1='60分钟') -> OrderedDict:
    """原始分钟K线结束，c 必须是基础周期的 CZSC 对象"""
    s = OrderedDict()
    k2, k3 = "K线", "结束"
    assert "分钟" in k1

    m = int(k1.replace("分钟", ""))
    dt: datetime = c.bars_raw[-1].dt
    if dt.minute % m == 0:
        v = "是"
    else:
        v = "否"

    signal = Signal(k1=k1, k2=k2, k3=k3, v1=v)
    s[signal.key] = signal.value
    return s


def get_s_op_time_span(c: CZSC, op: str = '开多', time_span=("14:00", "14:50")) -> OrderedDict:
    """日内操作时间区间，c 必须是基础周期的 CZSC 对象"""
    s = OrderedDict()
    dt: datetime = c.bars_raw[-1].dt
    k1 = f"{op}时间范围"
    assert len(time_span) == 2
    k2, k3 = time_span

    if k2 <= dt.strftime("%H:%M") <= k3:
        v = "是"
    else:
        v = "否"

    signal = Signal(k1=k1, k2=k2, k3=k3, v1=v)
    s[signal.key] = signal.value
    return s


def get_s_zdt(c: CZSC, di=1) -> OrderedDict:
    """计算倒数第di根K线的涨跌停信息"""
    s = OrderedDict()
    k1 = str(c.freq.value)
    k2 = f"倒{di}K"
    k3 = "ZDT"
    if len(c.bars_raw) < di + 2:
        v1 = "非涨跌停"
    else:
        bar = c.bars_raw[-di]
        if bar.close == bar.high:
            v1 = "涨停"
        elif bar.close == bar.low:
            v1 = "跌停"
        else:
            v1 = "非涨跌停"

    v = Signal(k1=k1, k2=k2, k3=k3, v1=v1)
    s[v.key] = v.value
    return s



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2021/11/21 17:48
describe: 技术分析相关信号的计算
"""
import numpy as np
from collections import OrderedDict

from .. import analyze
from ..objects import Signal
from ..enum import Freq
try:
    from ..utils.ta1 import MACD, SMA
except:
    from ..utils.ta import MACD, SMA


def get_s_single_k(c: analyze.CZSC, di: int = 1) -> OrderedDict:
    """获取倒数第i根K线的单K信号"""
    if c.freq not in [Freq.D, Freq.W]:
        return OrderedDict()

    if len(c.bars_raw) < di:
        return OrderedDict()

    s = OrderedDict()
    freq: Freq = c.freq
    k1 = str(freq.value)
    default_signals = [
        Signal(k1=k1, k2=f"倒{di}K", k3="状态", v1="其他", v2='其他', v3='其他'),
    ]
    for signal in default_signals:
        s[signal.key] = signal.value

    k = c.bars_raw[-di]
    if k.close > k.open:
        v = Signal(k1=k1, k2=f"倒{di}K", k3="状态", v1="上涨")
    else:
        v = Signal(k1=k1, k2=f"倒{di}K", k3="状态", v1="下跌")
    s[v.key] = v.value
    return s


def get_s_three_k(c: analyze.CZSC, di: int = 1) -> OrderedDict:
    """倒数第i根K线的三K信号

    :param c: CZSC 对象
    :param di: 最近一根K线为倒数第i根
    :return: 信号字典
    """
    assert di >= 1
    freq: Freq = c.freq
    k1 = str(freq.value)
    k2 = f"倒{di}K"

    s = OrderedDict()
    v = Signal(k1=k1, k2=k2, k3="三K形态", v1="其他", v2='其他', v3='其他')
    s[v.key] = v.value

    if len(c.bars_ubi) < 3 + di:
        return s

    if di == 1:
        tri = c.bars_ubi[-3:]
    else:
        tri = c.bars_ubi[-3 - di + 1:-di + 1]

    if tri[0].high > tri[1].high < tri[2].high:
        v = Signal(k1=k1, k2=k2, k3="三K形态", v1="底分型")
    elif tri[0].high < tri[1].high < tri[2].high:
        v = Signal(k1=k1, k2=k2, k3="三K形态", v1="向上走")
    elif tri[0].high < tri[1].high > tri[2].high:
        v = Signal(k1=k1, k2=k2, k3="三K形态", v1="顶分型")
    elif tri[0].high > tri[1].high > tri[2].high:
        v = Signal(k1=k1, k2=k2, k3="三K形态", v1="向下走")
    else:
        v = None

    if v and "其他" not in v.value:
        s[v.key] = v.value

    return s


def get_s_macd(c: analyze.CZSC, di: int = 1) -> OrderedDict:
    """获取倒数第i根K线的MACD相关信号"""
    freq: Freq = c.freq
    s = OrderedDict()

    k1 = str(freq.value)
    k2 = f"倒{di}K"
    default_signals = [
        Signal(k1=k1, k2=k2, k3="DIF多空", v1="其他", v2='其他', v3='其他'),
        Signal(k1=k1, k2=k2, k3="DIF方向", v1="其他", v2='其他', v3='其他'),

        Signal(k1=k1, k2=k2, k3="DEA多空", v1="其他", v2='其他', v3='其他'),
        Signal(k1=k1, k2=k2, k3="DEA方向", v1="其他", v2='其他', v3='其他'),

        Signal(k1=k1, k2=k2, k3="MACD多空", v1="其他", v2='其他', v3='其他'),
        Signal(k1=k1, k2=k2, k3="MACD方向", v1="其他", v2='其他', v3='其他'),
    ]
    for signal in default_signals:
        s[signal.key] = signal.value

    if len(c.bars_raw) < 100:
        return s

    if di == 1:
        close = np.array([x.close for x in c.bars_raw[-100:]])
    else:
        close = np.array([x.close for x in c.bars_raw[-100-di+1:-di+1]])
    dif, dea, macd = MACD(close, fastperiod=12, slowperiod=26, signalperiod=9)

    # DIF 多空信号
    dif_base = sum([abs(dif[-2] - dif[-1]), abs(dif[-3] - dif[-2]), abs(dif[-4] - dif[-3])]) / 3
    if dif[-1] > dif_base:
        v = Signal(k1=k1, k2=k2, k3="DIF多空", v1="多头")
    elif dif[-1] < -dif_base:
        v = Signal(k1=k1, k2=k2, k3="DIF多空", v1="空头")
    else:
        v = Signal(k1=k1, k2=k2, k3="DIF多空", v1="模糊")
    s[v.key] = v.value

    if dif[-1] > dif[-2] > dif[-3]:
        v = Signal(k1=k1, k2=k2, k3="DIF方向", v1="向上")
    elif dif[-1] < dif[-2] < dif[-3]:
        v = Signal(k1=k1, k2=k2, k3="DIF方向", v1="向下")
    else:
        v = Signal(k1=k1, k2=k2, k3="DIF方向", v1="模糊")
    s[v.key] = v.value

    # DEA 多空信号
    dea_base = sum([abs(dea[-2] - dea[-1]), abs(dea[-3] - dea[-2]), abs(dea[-4] - dea[-3])]) / 3
    if dea[-1] > dea_base:
        v = Signal(k1=k1, k2=k2, k3="DEA多空", v1="多头")
    elif dea[-1] < -dea_base:
        v = Signal(k1=k1, k2=k2, k3="DEA多空", v1="空头")
    else:
        v = Signal(k1=k1, k2=k2, k3="DEA多空", v1="模糊")
    s[v.key] = v.value

    # DEA 方向信号
    if dea[-1] > dea[-2]:
        v = Signal(k1=k1, k2=k2, k3="DEA方向", v1="向上")
    elif dea[-1] < dea[-2]:
        v = Signal(k1=k1, k2=k2, k3="DEA方向", v1="向下")
    else:
        v = Signal(k1=k1, k2=k2, k3="DEA方向", v1="模糊")
    s[v.key] = v.value

    # MACD 多空信号
    if macd[-1] >= 0:
        v = Signal(k1=k1, k2=k2, k3="MACD多空", v1="多头")
    else:
        v = Signal(k1=k1, k2=k2, k3="MACD多空", v1="空头")
    s[v.key] = v.value

    # MACD 方向信号
    if macd[-1] > macd[-2] > macd[-3]:
        v = Signal(k1=k1, k2=k2, k3="MACD方向", v1="向上")
    elif macd[-1] < macd[-2] < macd[-3]:
        v = Signal(k1=k1, k2=k2, k3="MACD方向", v1="向下")
    else:
        v = Signal(k1=k1, k2=k2, k3="MACD方向", v1="模糊")
    s[v.key] = v.value
    return s


def get_s_sma(c: analyze.CZSC, di: int = 1, t_seq=(5, 10, 20, 60)) -> OrderedDict:
    """获取倒数第i根K线的SMA相关信号"""
    freq: Freq = c.freq
    s = OrderedDict()

    k1 = str(freq.value)
    k2 = f"倒{di}K"
    for t in t_seq:
        x1 = Signal(k1=k1, k2=k2, k3=f"SMA{t}多空", v1="其他", v2='其他', v3='其他')
        x2 = Signal(k1=k1, k2=k2, k3=f"SMA{t}方向", v1="其他", v2='其他', v3='其他')
        s[x1.key] = x1.value
        s[x2.key] = x2.value

    n = max(t_seq) + 10
    if len(c.bars_raw) < n:
        return s

    if di == 1:
        close = np.array([x.close for x in c.bars_raw[-n:]])
    else:
        close = np.array([x.close for x in c.bars_raw[-n-di+1:-di+1]])

    for t in t_seq:
        sma = SMA(close, timeperiod=t)
        if close[-1] >= sma[-1]:
            v1 = Signal(k1=k1, k2=k2, k3=f"SMA{t}多空", v1="多头")
        else:
            v1 = Signal(k1=k1, k2=k2, k3=f"SMA{t}多空", v1="空头")
        s[v1.key] = v1.value

        if sma[-1] >= sma[-2]:
            v2 = Signal(k1=k1, k2=k2, k3=f"SMA{t}方向", v1="向上")
        else:
            v2 = Signal(k1=k1, k2=k2, k3=f"SMA{t}方向", v1="向下")
        s[v2.key] = v2.value
    return s





# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2021/11/7 17:09
describe: 支持分批买入卖出的高级交易员
"""
import os
import webbrowser
from typing import Callable, List
from pyecharts.charts import Tab
from pyecharts.components import Table
from pyecharts.options import ComponentTitleOpts

from ..analyze import CZSC, signals_counter
from ..objects import PositionLong, PositionShort, Operate, Event, RawBar
from ..utils.bar_generator import BarGenerator
from ..utils.cache import home_path
from .. import envs


class CzscAdvancedTrader:
    """缠中说禅技术分析理论之多级别联立交易决策类（支持分批开平仓 / 支持从任意周期开始交易）"""

    def __init__(self,
                 bg: BarGenerator,
                 get_signals: Callable,
                 long_events: List[Event] = None,
                 long_pos: PositionLong = None,
                 short_events: List[Event] = None,
                 short_pos: PositionShort = None,
                 max_bi_count: int = 50,
                 signals_n: int = 0,
                 ):
        """

        :param bg: K线合成器
        :param get_signals: 自定义的单级别信号计算函数
        :param long_events: 自定义的多头交易事件组合，推荐平仓事件放到前面
        :param long_pos: 多头仓位对象
        :param short_events: 自定义的空头交易事件组合，推荐平仓事件放到前面
        :param short_pos: 空头仓位对象
        :param max_bi_count: 单个级别最大保存笔的数量
        :param signals_n: 见 `CZSC` 对象
        """
        self.name = "CzscAdvancedTrader"
        self.bg = bg
        self.symbol = bg.symbol
        self.base_freq = bg.base_freq
        self.freqs = list(bg.bars.keys())
        self.get_signals = get_signals
        self.long_events = long_events
        self.long_pos = long_pos
        self.short_events = short_events
        self.short_pos = short_pos
        self.signals_n = signals_n
        self.signals_list = []
        self.verbose = envs.get_verbose()
        self.kas = {freq: CZSC(b, max_bi_count) for freq, b in bg.bars.items()}

        last_bar = self.kas[self.base_freq].bars_raw[-1]
        self.end_dt, self.bid, self.latest_price = last_bar.dt, last_bar.id, last_bar.close
        self.s = self.get_signals(self)

    def __repr__(self):
        return "<{} for {}>".format(self.name, self.symbol)

    def take_snapshot(self, file_html=None, width: str = "1400px", height: str = "580px"):
        """获取快照

        :param file_html: 交易快照保存的 html 文件名
        :param width: 图表宽度
        :param height: 图表高度
        :return:
        """
        tab = Tab(page_title="{}@{}".format(self.symbol, self.end_dt.strftime("%Y-%m-%d %H:%M")))
        for freq in self.freqs:
            chart = self.kas[freq].to_echarts(width, height)
            tab.add(chart, freq)

        signals = {k: v for k, v in self.s.items() if len(k.split("_")) == 3}
        for freq in self.freqs:
            freq_signals = {k: signals[k] for k in signals.keys() if k.startswith("{}_".format(freq))}
            for k in freq_signals.keys():
                signals.pop(k)
            if len(freq_signals) <= 0:
                continue
            t1 = Table()
            t1.add(["名称", "数据"], [[k, v] for k, v in freq_signals.items()])
            t1.set_global_opts(title_opts=ComponentTitleOpts(title="缠中说禅信号表", subtitle=""))
            tab.add(t1, f"{freq}信号")

        if len(signals) > 0:
            t1 = Table()
            t1.add(["名称", "数据"], [[k, v] for k, v in signals.items()])
            t1.set_global_opts(title_opts=ComponentTitleOpts(title="缠中说禅信号表", subtitle=""))
            tab.add(t1, "其他信号")

        if file_html:
            tab.render(file_html)
        else:
            return tab

    def open_in_browser(self, width="1400px", height="580px"):
        """直接在浏览器中打开分析结果"""
        file_html = os.path.join(home_path, "temp_czsc_advanced_trader.html")
        self.take_snapshot(file_html, width, height)
        webbrowser.open(file_html)

    def update(self, bar: RawBar):
        """输入基础周期已完成K线，更新信号，更新仓位"""
        self.bg.update(bar)
        for freq, b in self.bg.bars.items():
            self.kas[freq].update(b[-1])

        self.symbol = bar.symbol
        last_bar = self.kas[self.base_freq].bars_raw[-1]
        self.end_dt, self.bid, self.latest_price = last_bar.dt, last_bar.id, last_bar.close
        dt, bid, price = self.end_dt, self.bid, self.latest_price
        self.s = self.get_signals(self)
        if self.signals_n > 0:
            self.signals_list.append(self.s)
            self.signals_list = self.signals_list[-self.signals_n:]
            self.s.update(signals_counter(self.signals_list))

        # 遍历 long_events，更新 long_pos
        if self.long_events:
            assert isinstance(self.long_pos, PositionLong), "long_events 必须配合 PositionLong 使用"

            op = Operate.HO
            op_desc = ""

            for event in self.long_events:
                m, f = event.is_match(self.s)
                if m:
                    op = event.operate
                    op_desc = f"{event.name}@{f}"
                    break

            self.long_pos.update(dt, op, price, bid, op_desc)

        # 遍历 short_events，更新 short_pos
        if self.short_events:
            assert isinstance(self.short_pos, PositionShort), "short_events 必须配合 PositionShort 使用"

            op = Operate.HO
            op_desc = ""

            for event in self.short_events:
                m, f = event.is_match(self.s)
                if m:
                    op = event.operate
                    op_desc = f"{event.name}@{f}"
                    break

            self.short_pos.update(dt, op, price, bid, op_desc)





# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2021/3/10 11:21
describe: 缠论分型、笔的识别
"""
import os
import webbrowser
import traceback
from typing import List, Callable
from collections import OrderedDict

from .enum import Mark, Direction
from .objects import BI, FX, RawBar, NewBar, Signal
from .utils.echarts_plot import kline_pro
from . import envs


def remove_include(k1: NewBar, k2: NewBar, k3: RawBar):
    """去除包含关系：输入三根k线，其中k1和k2为没有包含关系的K线，k3为原始K线"""
    if k1.high < k2.high:
        direction = Direction.Up
    elif k1.high > k2.high:
        direction = Direction.Down
    else:
        k4 = NewBar(symbol=k3.symbol, id=k3.id, freq=k3.freq, dt=k3.dt, open=k3.open,
                    close=k3.close, high=k3.high, low=k3.low, vol=k3.vol, elements=[k3])
        return False, k4

    # 判断 k2 和 k3 之间是否存在包含关系，有则处理
    if (k2.high <= k3.high and k2.low >= k3.low) or (k2.high >= k3.high and k2.low <= k3.low):
        if direction == Direction.Up:
            high = max(k2.high, k3.high)
            low = max(k2.low, k3.low)
            dt = k2.dt if k2.high > k3.high else k3.dt
        elif direction == Direction.Down:
            high = min(k2.high, k3.high)
            low = min(k2.low, k3.low)
            dt = k2.dt if k2.low < k3.low else k3.dt
        else:
            raise ValueError

        if k3.open > k3.close:
            open_ = high
            close = low
        else:
            open_ = low
            close = high
        vol = k2.vol + k3.vol
        # 这里有一个隐藏Bug，len(k2.elements) 在一些及其特殊的场景下会有超大的数量，具体问题还没找到；
        # 临时解决方案是直接限定len(k2.elements)<=100
        elements = [x for x in k2.elements[:100] if x.dt != k3.dt] + [k3]
        k4 = NewBar(symbol=k3.symbol, id=k2.id, freq=k2.freq, dt=dt, open=open_,
                    close=close, high=high, low=low, vol=vol, elements=elements)
        return True, k4
    else:
        k4 = NewBar(symbol=k3.symbol, id=k3.id, freq=k3.freq, dt=k3.dt, open=k3.open,
                    close=k3.close, high=k3.high, low=k3.low, vol=k3.vol, elements=[k3])
        return False, k4


def check_fx(k1: NewBar, k2: NewBar, k3: NewBar):
    """查找分型"""
    fx = None
    if k1.high < k2.high > k3.high and k1.low < k2.low > k3.low:
        fx = FX(symbol=k1.symbol, dt=k2.dt, mark=Mark.G, high=k2.high,
                low=k2.low, fx=k2.high, elements=[k1, k2, k3])

    if k1.low > k2.low < k3.low and k1.high > k2.high < k3.high:
        fx = FX(symbol=k1.symbol, dt=k2.dt, mark=Mark.D, high=k2.high,
                low=k2.low, fx=k2.low, elements=[k1, k2, k3])

    return fx


def check_fxs(bars: List[NewBar]) -> List[FX]:
    """输入一串无包含关系K线，查找其中所有分型"""
    fxs = []
    for i in range(1, len(bars)-1):
        fx: FX = check_fx(bars[i-1], bars[i], bars[i+1])
        if isinstance(fx, FX):
            # 这里可能隐含Bug，默认情况下，fxs本身是顶底交替的，但是对于一些特殊情况下不是这样，这是不对的。
            # 临时处理方案，强制要求fxs序列顶底交替
            if len(fxs) >= 2 and fx.mark == fxs[-1].mark:
                fxs.pop()
            fxs.append(fx)
    return fxs


def check_bi(bars: List[NewBar]):
    """输入一串无包含关系K线，查找其中的一笔

    :param bars: 无包含关系K线列表
    :return:
    """
    min_bi_len = envs.get_min_bi_len()
    fxs = check_fxs(bars)
    if len(fxs) < 2:
        return None, bars

    fx_a = fxs[0]
    try:
        if fxs[0].mark == Mark.D:
            direction = Direction.Up
            fxs_b = [x for x in fxs if x.mark == Mark.G and x.dt > fx_a.dt and x.fx > fx_a.fx]
            if not fxs_b:
                return None, bars

            fx_b = fxs_b[0]
            for fx in fxs_b:
                if fx.high >= fx_b.high:
                    fx_b = fx

        elif fxs[0].mark == Mark.G:
            direction = Direction.Down
            fxs_b = [x for x in fxs if x.mark == Mark.D and x.dt > fx_a.dt and x.fx < fx_a.fx]
            if not fxs_b:
                return None, bars

            fx_b = fxs_b[0]
            for fx in fxs_b[1:]:
                if fx.low <= fx_b.low:
                    fx_b = fx
        else:
            raise ValueError
    except:
        traceback.print_exc()
        return None, bars

    bars_a = [x for x in bars if fx_a.elements[0].dt <= x.dt <= fx_b.elements[2].dt]
    bars_b = [x for x in bars if x.dt >= fx_b.elements[0].dt]

    # 判断fx_a和fx_b价格区间是否存在包含关系
    ab_include = (fx_a.high > fx_b.high and fx_a.low < fx_b.low) \
                 or (fx_a.high < fx_b.high and fx_a.low > fx_b.low)
    if len(bars_a) >= min_bi_len and not ab_include:
        fxs_ = [x for x in fxs if fx_a.elements[0].dt <= x.dt <= fx_b.elements[2].dt]
        bi = BI(symbol=fx_a.symbol, fx_a=fx_a, fx_b=fx_b, fxs=fxs_, direction=direction, bars=bars_a)

        low_ubi = min([x.low for x in bars_b])
        high_ubi = max([x.high for x in bars_b])
        if (bi.direction == Direction.Up and high_ubi > bi.high) \
                or (bi.direction == Direction.Down and low_ubi < bi.low):
            return None, bars
        else:
            return bi, bars_b
    else:
        return None, bars


def signals_counter(signals_list) -> OrderedDict:
    """信号连续出现次数记录

    :param signals_list: 存储的信号列表
        数据样例：
            signals_list = [
                {"dt": '2020-01-01', "日线_倒1K_SMA5多空": "空头_任意_任意_0"},
                {"dt": '2020-01-02', "日线_倒1K_SMA5多空": "多头_任意_任意_0"},
                {"dt": '2020-01-03', "日线_倒1K_SMA5多空": "多头_任意_任意_0"},
                {"dt": '2020-01-04', "日线_倒1K_SMA5多空": "多头_任意_任意_0"},
                {"dt": '2020-01-05', "日线_倒1K_SMA5多空": "多头_任意_任意_0"},
                {"dt": '2020-01-06', "日线_倒1K_SMA5多空": "多头_任意_任意_0"},
            ]
    :return: 信号连续次数
    """
    if not signals_list:
        return OrderedDict()

    signals = [Signal(f"{k}_{v}") for k, v in signals_list[-1].items()
               if len(k.split("_")) == 3 and "连续次数" not in k]

    s = OrderedDict()
    for signal in signals:
        k1 = signal.k1
        k2 = f"{signal.k2}#{signal.k3}"
        k3 = "连续次数"
        seq = [signal.is_match(x) for x in signals_list]
        assert seq[-1], "最后一个信号匹配结果必须为 True"

        n = 0
        for x in seq:
            if x:
                n += 1
            else:
                n = 0
        assert n >= 1, "连续次数小于1，不合逻辑"

        signal_c = Signal(k1=k1, k2=k2, k3=k3, v1=f"{n}次")
        s[signal_c.key] = signal_c.value

    return s


class CZSC:
    def __init__(self,
                 bars: List[RawBar],
                 max_bi_count: int = 50,
                 get_signals: Callable = None,
                 signals_n: int = 0):
        """

        :param bars: K线数据
        :param get_signals: 自定义的信号计算函数
        :param signals_n: 缓存n个历史时刻的信号，0 表示不缓存；缓存的数据，主要用于计算信号连续次数
        :param max_bi_count: 最大保存的笔数量
            默认值为 50，仅使用内置的信号和因子，不需要调整这个参数。
            如果进行新的信号计算需要用到更多的笔，可以适当调大这个参数。
        """
        self.verbose = envs.get_verbose()
        self.max_bi_count = max_bi_count
        self.signals_n = signals_n
        self.bars_raw = []  # 原始K线序列
        self.bars_ubi = []  # 未完成笔的无包含K线序列
        self.bi_list: List[BI] = []
        self.symbol = bars[0].symbol
        self.freq = bars[0].freq
        self.get_signals = get_signals
        self.signals = None
        self.signals_list = []

        for bar in bars:
            self.update(bar)

    def __repr__(self):
        return "<CZSC~{}~{}>".format(self.symbol, self.freq.value)

    def __update_bi(self):
        bars_ubi = self.bars_ubi
        if len(bars_ubi) < 3:
            return

        # 查找笔
        if not self.bi_list:
            # 第一个笔的查找
            fxs = check_fxs(bars_ubi)
            if not fxs:
                return

            fx_a = fxs[0]
            fxs_a = [x for x in fxs if x.mark == fx_a.mark]
            for fx in fxs_a:
                if (fx_a.mark == Mark.D and fx.low <= fx_a.low) \
                        or (fx_a.mark == Mark.G and fx.high >= fx_a.high):
                    fx_a = fx
            bars_ubi = [x for x in bars_ubi if x.dt >= fx_a.elements[0].dt]

            bi, bars_ubi_ = check_bi(bars_ubi)
            if isinstance(bi, BI):
                self.bi_list.append(bi)
            self.bars_ubi = bars_ubi_
            return

        last_bi = self.bi_list[-1]

        # 如果上一笔被破坏，将上一笔的bars与bars_ubi进行合并
        if (last_bi.direction == Direction.Up and bars_ubi[-1].high > last_bi.high) \
                or (last_bi.direction == Direction.Down and bars_ubi[-1].low < last_bi.low):
            bars_ubi_a = last_bi.bars[:-1] + [x for x in bars_ubi if x.dt >= last_bi.bars[-1].dt]
            self.bi_list.pop(-1)
        else:
            bars_ubi_a = bars_ubi

        if self.verbose and len(bars_ubi_a) > 100:
            print(f"czsc_update_bi: {self.symbol} - {self.freq} - {bars_ubi_a[-1].dt} 未完成笔延伸数量: {len(bars_ubi_a)}")

        bi, bars_ubi_ = check_bi(bars_ubi_a)
        self.bars_ubi = bars_ubi_
        if isinstance(bi, BI):
            self.bi_list.append(bi)

    def update(self, bar: RawBar):
        """更新分析结果

        :param bar: 单根K线对象
        """
        # 更新K线序列
        if not self.bars_raw or bar.dt != self.bars_raw[-1].dt:
            self.bars_raw.append(bar)
            last_bars = [bar]
        else:
            self.bars_raw[-1] = bar
            last_bars = self.bars_ubi[-1].elements
            last_bars[-1] = bar
            self.bars_ubi.pop(-1)

        # 去除包含关系
        bars_ubi = self.bars_ubi
        for bar in last_bars:
            if len(bars_ubi) < 2:
                bars_ubi.append(NewBar(symbol=bar.symbol, id=bar.id, freq=bar.freq, dt=bar.dt,
                                       open=bar.open, close=bar.close,
                                       high=bar.high, low=bar.low, vol=bar.vol, elements=[bar]))
            else:
                k1, k2 = bars_ubi[-2:]
                has_include, k3 = remove_include(k1, k2, bar)
                if has_include:
                    bars_ubi[-1] = k3
                else:
                    bars_ubi.append(k3)
        self.bars_ubi = bars_ubi

        # 更新笔
        self.__update_bi()
        self.bi_list = self.bi_list[-self.max_bi_count:]
        if self.bi_list:
            sdt = self.bi_list[0].fx_a.elements[0].dt
            s_index = 0
            for i, bar in enumerate(self.bars_raw):
                if bar.dt >= sdt:
                    s_index = i
                    break
            self.bars_raw = self.bars_raw[s_index:]

        if self.get_signals:
            self.signals = self.get_signals(c=self)
            if self.signals_n > 0:
                self.signals_list.append(self.signals)
                self.signals_list = self.signals_list[-self.signals_n:]
                self.signals.update(signals_counter(self.signals_list))
        else:
            self.signals = OrderedDict()

    def to_echarts(self, width: str = "1400px", height: str = '580px'):
        kline = [x.__dict__ for x in self.bars_raw]
        if len(self.bi_list) > 0:
            bi = [{'dt': x.fx_a.dt, "bi": x.fx_a.fx} for x in self.bi_list] + \
                 [{'dt': self.bi_list[-1].fx_b.dt, "bi": self.bi_list[-1].fx_b.fx}]
            fx = []
            for bi_ in self.bi_list:
                fx.extend([{'dt': x.dt, "fx": x.fx} for x in bi_.fxs[1:]])
        else:
            bi = None
            fx = None
        chart = kline_pro(kline, bi=bi, fx=fx, width=width, height=height,
                          title="{}-{}".format(self.symbol, self.freq.value))
        return chart

    def open_in_browser(self, width: str = "1400px", height: str = '580px'):
        """直接在浏览器中打开分析结果

        :param width: 图表宽度
        :param height: 图表高度
        :return:
        """
        home_path = os.path.expanduser("~")
        file_html = os.path.join(home_path, "temp_czsc.html")
        chart = self.to_echarts(width, height)
        chart.render(file_html)
        webbrowser.open(file_html)

    @property
    def last_bi_extend(self):
        """判断最后一笔是否在延伸中，True 表示延伸中"""
        if self.bi_list[-1].direction == Direction.Up \
                and max([x.high for x in self.bars_ubi]) > self.bi_list[-1].high:
            return True

        if self.bi_list[-1].direction == Direction.Down \
                and min([x.low for x in self.bars_ubi]) < self.bi_list[-1].low:
            return True

        return False

    @property
    def finished_bis(self) -> List[BI]:
        """返回当下基本确认完成的笔列表"""
        if not self.bi_list:
            return []
        else:
            if self.last_bi_extend:
                return self.bi_list[:-1]
        return self.bi_list





# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2021/12/13 17:39
describe: 事件性能分析
"""
import os
import os.path
import traceback
import pandas as pd
import matplotlib.pyplot as plt
from deprecated import deprecated
from datetime import timedelta, datetime
from tqdm import tqdm
from typing import Callable, List
from czsc.objects import Factor
from czsc.data.ts_cache import TsDataCache
from czsc.sensors.utils import generate_signals
from czsc.utils import io
from czsc.utils import WordWriter


plt.style.use('ggplot')
plt.rcParams['font.sans-serif'] = ['SimHei']
plt.rcParams['axes.unicode_minus'] = False


@deprecated(reason="使用 SignalsPerformance 替代信号分析", version='0.9.0')
class FactorsSensor:
    """因子（Factor）感应器：分析各种信号和因子的表现"""

    def __init__(self,
                 results_path: str,
                 sdt: str,
                 edt: str,
                 dc: TsDataCache,
                 base_freq: str,
                 freqs: List[str],
                 get_signals: Callable,
                 get_factors: Callable):

        self.name = self.__class__.__name__
        self.version = "V20211213"
        os.makedirs(results_path, exist_ok=True)
        self.results_path = results_path
        self.sdt = sdt
        self.edt = edt

        self.get_signals = get_signals
        self.get_factors = get_factors
        self.factors: List[Factor] = get_factors()
        self.base_freq = base_freq
        self.freqs = freqs
        self.file_docx = os.path.join(results_path, f'factors_sensor_{sdt}_{edt}.docx')
        self.writer = WordWriter(self.file_docx)

        self.dc = dc
        self.betas = ['000001.SH', '000016.SH', '000905.SH', '000300.SH', '399001.SZ', '399006.SZ']

        self.file_sf = os.path.join(results_path, f'factors_{sdt}_{edt}.pkl')
        self.signals_path = os.path.join(results_path, 'signals')
        os.makedirs(self.signals_path, exist_ok=True)
        if os.path.exists(self.file_sf):
            self.sf = io.read_pkl(self.file_sf)
        else:
            self.sf = self.get_stock_factors()
            io.save_pkl(self.sf, self.file_sf)

    def get_share_factors(self, ts_code: str, name: str):
        """获取单个标的因子信息"""
        dc = self.dc
        sdt = self.sdt
        edt = self.edt
        factors = self.factors

        start_date = pd.to_datetime(self.sdt) - timedelta(days=3000)
        bars = dc.pro_bar(ts_code=ts_code, start_date=start_date, end_date=edt, freq='D', asset="E", raw_bar=True)
        n_bars = dc.pro_bar(ts_code=ts_code, start_date=sdt, end_date=edt, freq='D', asset="E", raw_bar=False)
        nb_dicts = {row['trade_date'].strftime("%Y%m%d"): row for row in n_bars.to_dict("records")}
        signals = generate_signals(bars, sdt, self.base_freq, self.freqs, self.get_signals)

        results = []
        for s in signals:
            row = {'name': name, 'ts_code': ts_code}
            for factor in factors:
                row[factor.name] = factor.is_match(s)

            nb_info = nb_dicts.get(s['dt'].strftime("%Y%m%d"), None)
            row.update(nb_info)
            results.append(row)

        df_res = pd.DataFrame(results)
        if df_res.empty:
            return df_res

        df_res = df_res[pd.to_datetime(sdt) <= df_res['trade_date']]
        df_res = df_res[df_res['trade_date'] <= pd.to_datetime(edt)]

        # 加入总市值
        df_ = dc.daily_basic(ts_code, sdt, dc.edt)
        df_['trade_date'] = pd.to_datetime(df_['trade_date'])
        df_res = df_res.merge(df_[['trade_date', 'total_mv']], on='trade_date', how='left')
        return signals, df_res

    def get_stock_factors(self):
        """获取全部股票的因子信息"""
        stocks = self.dc.stock_basic()

        all_factors = []
        for row in tqdm(stocks.to_dict('records'), desc="get_stock_factors"):
            ts_code = row['ts_code']
            name = row['name']
            try:
                signals, factors = self.get_share_factors(ts_code, name)
                all_factors.append(factors)
                file_signals = os.path.join(self.signals_path, f'{ts_code}.pkl')
                io.save_pkl(signals, file_signals)
            except:
                print(f"get_share_factors error: {ts_code}, {name}")
                traceback.print_exc()

        df_factors = pd.concat(all_factors, ignore_index=True)
        return df_factors

    def validate_performance(self):
        factors = self.factors
        sf = self.sf
        results = [{
            "name": "全市场", "count": len(sf), 'n1b': sf.n1b.mean(), 'n2b': sf.n2b.mean(),
            'n3b': sf.n3b.mean(), 'n5b': sf.n5b.mean(), 'n10b': sf.n10b.mean(), 'n20b': sf.n20b.mean()
        }]

        for factor in factors:
            df = sf[sf[factor.name]]
            row = {"name": factor.name, "count": len(df)}
            row.update(df[['n1b', 'n2b', 'n3b', 'n5b', 'n10b', 'n20b']].mean().to_dict())
            results.append(row)
        df_nb_info = pd.DataFrame(results)
        df_nb_info.to_excel(os.path.join(self.results_path, f"factors_nb_info.xlsx"), index=False)



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2022/4/13 15:03
describe: 持仓相关信号
"""
from collections import OrderedDict
from czsc.objects import Operate, Signal, PositionLong
from czsc.traders.advanced import CzscAdvancedTrader


def get_s_long01(cat: CzscAdvancedTrader, th=300):
    """多头持仓信号：亏损"""
    pos = cat.long_pos
    k1, k2, k3 = '多头', '亏损', f'超{th}BP'
    s = OrderedDict()
    v1 = '否'
    if pos.pos > 0:
        latest_price = cat.latest_price
        last_o = [x for x in pos.operates[-50:] if x['op'] == Operate.LO][-1]
        last_o_price = last_o['price']
        yk = (latest_price - last_o_price) / last_o_price
        if yk * 10000 < -th:
            v1 = '是'

    signal = Signal(k1=k1, k2=k2, k3=k3, v1=v1)
    s[signal.key] = signal.value
    return s


def get_s_long02(cat: CzscAdvancedTrader, th=300):
    """多头持仓信号：回撤"""
    pos = cat.long_pos
    k1, k2, k3 = '多头', '回撤', f'超{th}BP'
    s = OrderedDict()
    v1 = '否'
    if pos.pos > 0 and cat.latest_price <= pos.long_high:
        hc = abs(cat.latest_price / pos.long_high - 1)
        if hc * 10000 > th:
            v1 = '是'

    signal = Signal(k1=k1, k2=k2, k3=k3, v1=v1)
    s[signal.key] = signal.value
    return s


def get_s_long03(cat: CzscAdvancedTrader, th=100):
    """多头持仓信号：持仓时间"""
    pos = cat.long_pos
    k1, k2, k3 = '多头', '持仓时间', f'超{th}根基础K线'
    s = OrderedDict()
    v1 = '否'
    if pos.pos > 0:
        lo_bid = [x for x in pos.operates[-50:] if x['op'] == Operate.LO][-1]['bid']
        bid = cat.bg.bars[cat.base_freq][-1].id
        if bid - lo_bid > th:
            v1 = '是'

    signal = Signal(k1=k1, k2=k2, k3=k3, v1=v1)
    s[signal.key] = signal.value
    return s


def get_s_long04(cat: CzscAdvancedTrader, th=5):
    """多头持仓信号：持仓亏损时间"""
    pos = cat.long_pos
    k1, k2, k3 = '多头', '持仓亏损时间', f'超{th}根基础K线'
    s = OrderedDict()
    v1 = '否'
    if pos.pos > 0:
        lo = [x for x in pos.operates[-50:] if x['op'] == Operate.LO][-1]
        hold_bars = [x for x in cat.kas[cat.base_freq].bars_raw[-(th+50):] if x.id >= lo['bid']]
        loss_bars = [1 if x.close < lo['price'] else 0 for x in hold_bars]
        if sum(loss_bars) > th:
            v1 = '是'

    signal = Signal(k1=k1, k2=k2, k3=k3, v1=v1)
    s[signal.key] = signal.value
    return s


def get_s_long05(cat: CzscAdvancedTrader, span="月", th=500):
    """多头持仓信号：周期累计亏损"""
    pos: PositionLong = cat.long_pos
    k1, k2, k3 = '多头', f'本{span}累计亏损', f'超{th}BP'
    s = OrderedDict()

    if span == '周':
        dt_fmt = '%Y年%W周'
    elif span == '月':
        dt_fmt = '%Y年%m月'
    else:
        raise ValueError
    dt_ = cat.end_dt.strftime(dt_fmt)
    pairs = [x for x in pos.pairs if x['平仓时间'].strftime(dt_fmt) == dt_]

    v1 = '否'
    if sum([x['盈亏比例'] for x in pairs]) < -th / 10000:
        v1 = '是'

    signal = Signal(k1=k1, k2=k2, k3=k3, v1=v1)
    s[signal.key] = signal.value
    return s



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2022/4/4 22:43
describe: 以 CzscAdvancedTrader 作为输入的信号计算
"""
from typing import List, Union
from collections import OrderedDict
from czsc.objects import Freq, Operate, Direction, BI, FakeBI, Signal, PositionLong, PositionShort
from czsc.traders.advanced import CzscAdvancedTrader


def get_s_position(cat: CzscAdvancedTrader, pos: [PositionLong, PositionShort]):
    """计算持仓信号

    :return:
    """
    if isinstance(pos, PositionLong):
        k1 = "多头"
    elif isinstance(pos, PositionShort):
        k1 = "空头"
    else:
        raise ValueError

    s = OrderedDict()
    default_signals = [
        Signal(k1=k1, k2="最大", k3='盈利', v1="其他", v2="其他", v3="其他"),
        Signal(k1=k1, k2="最大", k3='回撤', v1="其他", v2="其他", v3="其他"),
        Signal(k1=k1, k2="最大", k3='回撤盈利比', v1="其他", v2="其他", v3="其他"),

        Signal(k1=k1, k2="累计", k3='盈亏', v1="其他", v2="其他", v3="其他"),
        Signal(k1=k1, k2="持仓", k3='时间', v1="其他", v2="其他", v3="其他"),
        Signal(k1=k1, k2="持仓", k3='基础K线数量', v1="其他", v2="其他", v3="其他"),
    ]
    for signal_ in default_signals:
        s[signal_.key] = signal_.value

    if pos.pos == 0:
        return s

    base_freq = cat.base_freq
    latest_price = cat.latest_price
    bid = cat.bg.bars[base_freq][-1].id
    end_dt = cat.bg.bars[base_freq][-1].dt

    if isinstance(pos, PositionLong):
        last_o = [x for x in pos.operates[-50:] if x['op'] == Operate.LO][-1]
        last_o_price = last_o['price']
        yl = pos.long_high / last_o_price - 1                   # 最大盈利
        hc = abs(latest_price / pos.long_high - 1)              # 最大回撤
        yk = (latest_price - last_o_price) / last_o_price       # 累计盈亏
    else:
        last_o = [x for x in pos.operates[-50:] if x['op'] == Operate.SO][-1]
        last_o_price = last_o['price']
        yl = last_o_price / pos.short_low - 1                   # 最大盈利
        hc = abs(pos.short_low / latest_price - 1)              # 最大回撤
        yk = (last_o_price - latest_price) / last_o_price       # 累计盈亏

    last_o_dt = last_o['dt']
    last_o_bid = last_o['bid']

    hc_yl_rate = hc / (yl + 0.000001)                   # 最大回撤盈利比
    hold_time = (end_dt - last_o_dt).total_seconds()   # 持仓时间，单位：秒
    hold_nbar = bid - last_o_bid                       # 持仓基础K线数量
    assert yl >= 0 and hc >= 0 and hc_yl_rate >= 0

    # ----------------------------------------------------------------------------------
    if yl > 0.15:
        v1 = "超过1500BP"
    elif yl > 0.1:
        v1 = "超过1000BP"
    elif yl > 0.08:
        v1 = "超过800BP"
    elif yl > 0.05:
        v1 = "超过500BP"
    elif yl > 0.03:
        v1 = "超过300BP"
    else:
        v1 = "低于300BP"
    v = Signal(k1=k1, k2="最大", k3='盈利', v1=v1)
    s[v.key] = v.value

    # ----------------------------------------------------------------------------------
    if hc > 0.15:
        v1 = "超过1500BP"
    elif hc > 0.1:
        v1 = "超过1000BP"
    elif hc > 0.08:
        v1 = "超过800BP"
    elif hc > 0.05:
        v1 = "超过500BP"
    elif hc > 0.03:
        v1 = "超过300BP"
    else:
        v1 = "低于300BP"
    v = Signal(k1=k1, k2="最大", k3='回撤', v1=v1)
    s[v.key] = v.value

    # ----------------------------------------------------------------------------------
    if hc_yl_rate > 0.8:
        v1 = "大于08"
    elif hc_yl_rate > 0.6:
        v1 = "大于06"
    elif hc_yl_rate > 0.5:
        v1 = "大于05"
    elif hc_yl_rate > 0.3:
        v1 = "大于03"
    else:
        v1 = "小于03"
    v = Signal(k1=k1, k2="最大", k3='回撤盈利比', v1=v1)
    s[v.key] = v.value

    # ----------------------------------------------------------------------------------
    if yk >= 0:
        v1 = "盈利"
    else:
        v1 = "亏损"

    if abs(yk) > 0.15:
        v2 = "超过1500BP"
    elif abs(yk) > 0.1:
        v2 = "超过1000BP"
    elif abs(yk) > 0.08:
        v2 = "超过800BP"
    elif abs(yk) > 0.05:
        v2 = "超过500BP"
    elif abs(yk) > 0.03:
        v2 = "超过300BP"
    else:
        v2 = "低于300BP"
    v = Signal(k1=k1, k2="累计", k3='盈亏', v1=v1, v2=v2)
    s[v.key] = v.value

    # ----------------------------------------------------------------------------------
    if hold_time > 3600 * 24 * 13:
        v1 = "超过13天"
    elif hold_time > 3600 * 24 * 8:
        v1 = "超过8天"
    elif hold_time > 3600 * 24 * 5:
        v1 = "超过5天"
    elif hold_time > 3600 * 24 * 3:
        v1 = "超过3天"
    else:
        v1 = "低于3天"
    v = Signal(k1=k1, k2="持仓", k3='时间', v1=v1)
    s[v.key] = v.value

    # ----------------------------------------------------------------------------------
    if hold_nbar > 300:
        v1 = "超过300根"
    elif hold_nbar > 200:
        v1 = "超过200根"
    elif hold_nbar > 150:
        v1 = "超过150根"
    elif hold_nbar > 100:
        v1 = "超过100根"
    elif hold_nbar > 50:
        v1 = "超过50根"
    else:
        v1 = "低于50根"
    v = Signal(k1=k1, k2="持仓", k3='基础K线数量', v1=v1)
    s[v.key] = v.value

    return s






# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2019/10/29 15:01
"""
import os
import shutil

from . import envs
from . import ai
from . import utils
from . import traders
from . import sensors
from . import aphorism
from .analyze import CZSC
from .traders.advanced import CzscAdvancedTrader
from .utils.ta import SMA, EMA, MACD, KDJ
from .objects import Freq, Operate, Direction, Signal, Factor, Event, RawBar, NewBar

__version__ = "0.8.23"
__author__ = "zengbin93"
__email__ = "zeng_bin8888@163.com"
__date__ = "20220414"

print(f"欢迎使用CZSC！当前版本标识为 {__version__}@{__date__}\n")
aphorism.print_one()

home_path = os.path.join(os.path.expanduser("~"), '.czsc')
os.makedirs(home_path, exist_ok=True)


def get_dir_size(path):
    """获取目录大小，单位：Bytes"""
    total = 0
    with os.scandir(path) as it:
        for entry in it:
            if entry.is_file():
                total += entry.stat().st_size
            elif entry.is_dir():
                total += get_dir_size(entry.path)
    return total


if get_dir_size(home_path) > pow(1024, 3) and envs.get_verbose():
    print(f"{home_path} 目录缓存超过1GB，请适当清理。调用 empty_cache_path 可以直接清空缓存")


def empty_cache_path():
    shutil.rmtree(home_path)
    os.makedirs(home_path, exist_ok=False)
    print(f"已清空缓存文件夹：{home_path}")




# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2021/11/21 17:48
describe: 信号系统
"""
from . import bxt
from . import ta
from . import other
from . import vol
from . import cat
from . import pos




# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2021/11/7 21:07
"""
import pandas as pd
from tqdm import tqdm
from collections import OrderedDict
from czsc import signals
from czsc.traders.advanced import CzscAdvancedTrader, BarGenerator
from czsc.objects import Signal, Factor, Event, Operate, PositionLong, PositionShort
from test.test_analyze import read_1min, read_daily


def get_signals(cat: CzscAdvancedTrader) -> OrderedDict:
    s = OrderedDict({"symbol": cat.symbol, "dt": cat.end_dt, "close": cat.latest_price})
    for _, c in cat.kas.items():
        s.update(signals.bxt.get_s_like_bs(c, di=1))

    if cat.long_pos:
        s.update(signals.cat.get_s_position(cat, cat.long_pos))
    if cat.short_pos:
        s.update(signals.cat.get_s_position(cat, cat.short_pos))
    return s


def test_daily_trader():
    bars = read_daily()
    kg = BarGenerator(base_freq='日线', freqs=['周线', '月线'])
    for bar in bars[:1000]:
        kg.update(bar)

    ct = CzscAdvancedTrader(kg, get_signals)

    signals_ = []
    for bar in bars[1000:]:
        ct.update(bar)
        signals_.append(dict(ct.s))

    assert len(signals_) == 2332


def run_advanced_trader(T0=True):
    bars = read_1min()
    kg = BarGenerator(base_freq='1分钟', freqs=['5分钟', '15分钟', '30分钟', '60分钟', '日线'], max_count=3000)
    for row in tqdm(bars[:150000], desc='init kg'):
        kg.update(row)

    long_events = [
        Event(name="开多", operate=Operate.LO, factors=[
            Factor(name="5分钟一买", signals_all=[Signal("5分钟_倒1笔_类买卖点_类一买_任意_任意_0")]),
            Factor(name="1分钟一买", signals_all=[Signal("1分钟_倒1笔_类买卖点_类一买_任意_任意_0")]),
        ]),

        Event(name="加多1", operate=Operate.LA1, factors=[
            Factor(name="5分钟二买", signals_all=[Signal("5分钟_倒1笔_类买卖点_类二买_任意_任意_0")]),
            Factor(name="1分钟二买", signals_all=[Signal("1分钟_倒1笔_类买卖点_类二买_任意_任意_0")]),
        ]),

        Event(name="加多2", operate=Operate.LA1, factors=[
            Factor(name="5分钟三买", signals_all=[Signal("5分钟_倒1笔_类买卖点_类三买_任意_任意_0")]),
            Factor(name="1分钟三买", signals_all=[Signal("1分钟_倒1笔_类买卖点_类三买_任意_任意_0")]),
        ]),

        Event(name="平多", operate=Operate.LE, factors=[
            Factor(name="5分钟二卖", signals_all=[Signal("5分钟_倒1笔_类买卖点_类二卖_任意_任意_0")]),
            Factor(name="5分钟三卖", signals_all=[Signal("5分钟_倒1笔_类买卖点_类三卖_任意_任意_0")])
        ]),
    ]
    long_pos = PositionLong(symbol='000001.SH', hold_long_a=0.5, hold_long_b=0.8, hold_long_c=1, T0=T0)
    short_events = [
        Event(name="开空", operate=Operate.SO, factors=[
            Factor(name="5分钟一买", signals_all=[Signal("5分钟_倒1笔_类买卖点_类一买_任意_任意_0")]),
            Factor(name="1分钟一买", signals_all=[Signal("1分钟_倒1笔_类买卖点_类一买_任意_任意_0")]),
        ]),

        Event(name="加空1", operate=Operate.SA1, factors=[
            Factor(name="5分钟二买", signals_all=[Signal("5分钟_倒1笔_类买卖点_类二买_任意_任意_0")]),
            Factor(name="1分钟二买", signals_all=[Signal("1分钟_倒1笔_类买卖点_类二买_任意_任意_0")]),
        ]),

        Event(name="加空2", operate=Operate.SA1, factors=[
            Factor(name="5分钟三买", signals_all=[Signal("5分钟_倒1笔_类买卖点_类三买_任意_任意_0")]),
            Factor(name="1分钟三买", signals_all=[Signal("1分钟_倒1笔_类买卖点_类三买_任意_任意_0")]),
        ]),

        Event(name="平空", operate=Operate.SE, factors=[
            Factor(name="5分钟二卖", signals_all=[Signal("5分钟_倒1笔_类买卖点_类二卖_任意_任意_0")]),
            Factor(name="5分钟三卖", signals_all=[Signal("5分钟_倒1笔_类买卖点_类三卖_任意_任意_0")])
        ]),
    ]
    short_pos = PositionShort(symbol='000001.SH', hold_short_a=0.5, hold_short_b=0.8, hold_short_c=1, T0=T0)
    ct = CzscAdvancedTrader(kg, get_signals, long_events=long_events, long_pos=long_pos,
                            short_events=short_events, short_pos=short_pos)
    assert len(ct.s) == 21
    for row in tqdm(bars[150000:], desc="trade"):
        ct.update(row)
        # if long_pos.pos_changed:
        #     print(" : long op     : ", long_pos.operates[-1])
        # if short_pos.pos_changed:
        #     print(" : short op    : ", short_pos.operates[-1])

        if ct.long_pos.pos > 0:
            assert ct.long_pos.long_high > 0
            assert ct.long_pos.long_cost > 0
            assert ct.long_pos.long_bid > 0

        if ct.short_pos.pos > 0:
            assert ct.short_pos.short_low > 0
            assert ct.short_pos.short_cost > 0
            assert ct.short_pos.short_bid > 0

    long_yk = pd.DataFrame(ct.long_pos.pairs)['盈亏比例'].sum()
    short_yk = pd.DataFrame(ct.short_pos.pairs)['盈亏比例'].sum()
    assert abs(long_yk) == abs(short_yk)
    print(f"\nT0={T0}: 多头累计盈亏比例：{long_yk}；空头累计盈亏比例：{short_yk}")
    if not T0:
        assert ct.s['多头_最大_盈利'] == '超过800BP_任意_任意_0'
        assert ct.s['多头_累计_盈亏'] == '盈利_超过800BP_任意_0'
        assert ct.s['空头_最大_回撤'] == '超过800BP_任意_任意_0'
        assert ct.s['空头_累计_盈亏'] == '亏损_超过800BP_任意_0'


def test_advanced_trader():
    run_advanced_trader(T0=False)
    run_advanced_trader(T0=True)





# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2022/2/14 17:25
describe: 基于 Tushare 分钟数据的择时策略快速回测
"""

import os
import inspect
import traceback
import pandas as pd
from tqdm import tqdm
from typing import Callable

from ..data.ts_cache import TsDataCache
from ..traders.utils import trader_fast_backtest, freq_cn2ts
from ..utils import x_round


def read_raw_results(raw_path, trade_dir="long"):
    """读入指定路径下的回测原始结果

    :param raw_path: 原始结果路径
    :param trade_dir: 交易方向
    :return:
    """
    assert trade_dir in ['long', 'short']

    ops, pairs, p = [], [], []
    for file in tqdm(os.listdir(raw_path)):
        if len(file) != 14:
            continue
        file = os.path.join(raw_path, file)
        try:
            ops.append(pd.read_excel(file, sheet_name=f'{trade_dir}_operates'))
            pairs.append(pd.read_excel(file, sheet_name=f'{trade_dir}_pairs'))
            p.append(pd.read_excel(file, sheet_name=f'{trade_dir}_performance'))
        except:
            print(f"read_raw_results: fail on {file}")

    df_pairs = pd.concat(pairs, ignore_index=True)
    df_ops = pd.concat(ops, ignore_index=True)
    df_p = pd.concat(p, ignore_index=True)
    return df_ops, df_pairs, df_p


class TraderPerformance:
    """择时交易效果评估"""

    def __init__(self, df_pairs: pd.DataFrame, ):
        """

        :param df_pairs: 全部交易对
        """
        time_convert = lambda x: (x.strftime("%Y年"), x.strftime("%Y年%m月"), x.strftime("%Y-%m-%d"),
                                  f"{x.year}年第{x.weekofyear}周" if x.weekofyear >= 10 else f"{x.year}年第0{x.weekofyear}周",
                                  )
        df_pairs[['开仓年', '开仓月', '开仓日', '开仓周']] = list(df_pairs['开仓时间'].apply(time_convert))
        df_pairs[['平仓年', '平仓月', '平仓日', '平仓周']] = list(df_pairs['平仓时间'].apply(time_convert))

        self.df_pairs = df_pairs
        # 指定哪些列可以用来进行聚合分析
        self.agg_columns = ['标的代码', '交易方向', '开仓年', '平仓年', '开仓月', '平仓月',
                            '开仓周', '平仓周', '开仓日', '平仓日']

    @staticmethod
    def get_pairs_statistics(df_pairs: pd.DataFrame):
        """统计一组交易的基本信息

        :param df_pairs:
        :return:
        """
        if len(df_pairs) == 0:
            info = {
                "交易标的数量": 0,
                "总体交易次数": 0,
                "平均持仓天数": 0,

                "平均单笔收益": 0,
                "最大单笔收益": 0,
                "最小单笔收益": 0,

                "交易胜率": 0,
                "累计盈亏比": 0,
                "交易得分": 0,
                "每自然日收益": 0,
            }
            return info

        win_pct = x_round(len(df_pairs[df_pairs['盈亏比例'] > 0]) / len(df_pairs), 4)
        df_gain = df_pairs[df_pairs['盈亏比例'] > 0]
        df_loss = df_pairs[df_pairs['盈亏比例'] <= 0]
        gain = df_gain['盈亏比例'].sum()
        loss = abs(df_loss['盈亏比例'].sum())

        # 限制累计盈亏比最大有效值
        gain_loss_rate = min(x_round(gain / (loss + 0.000001), 2), 5)

        info = {
            "交易标的数量": df_pairs['标的代码'].nunique(),
            "总体交易次数": len(df_pairs),
            "平均持仓天数": x_round(df_pairs['持仓天数'].mean(), 2),

            "平均单笔收益": x_round(df_pairs['盈亏比例'].mean() * 10000, 2),
            "最大单笔收益": x_round(df_pairs['盈亏比例'].max() * 10000, 2),
            "最小单笔收益": x_round(df_pairs['盈亏比例'].min() * 10000, 2),

            "交易胜率": win_pct,
            "累计盈亏比": gain_loss_rate,
            "交易得分": x_round(gain_loss_rate * win_pct, 4),
        }

        info['每自然日收益'] = x_round(info['平均单笔收益'] / info['平均持仓天数'], 2)
        return info

    def agg_statistics(self, col: str):
        """按列聚合进行交易对评价"""
        df_pairs = self.df_pairs.copy()
        assert col in self.agg_columns, f"{col} 不是支持聚合的列，参考：{self.agg_columns}"

        results = []
        for name, dfg in df_pairs.groupby(col):
            if dfg.empty:
                continue

            res = {col: name}
            res.update(self.get_pairs_statistics(dfg))
            results.append(res)
        df = pd.DataFrame(results)
        return df

    @property
    def basic_info(self):
        """写入基础信息"""
        df_pairs = self.df_pairs.copy()
        return self.get_pairs_statistics(df_pairs)

    def agg_to_excel(self, file_xlsx):
        """遍历聚合列，保存结果到 Excel 文件中"""
        f = pd.ExcelWriter(file_xlsx)
        for col in self.agg_columns:
            df_ = self.agg_statistics(col)
            df_.to_excel(f, sheet_name=f"{col}聚合", index=False)
        f.close()
        print(f"聚合分析结果文件：{file_xlsx}")


class TsStocksBacktest:
    """基于 Tushare 数据的择时回测系统（股票市场）"""

    def __init__(self,
                 dc: TsDataCache,
                 strategy: Callable,
                 init_n: int,
                 sdt: str,
                 edt: str,
                 ):
        """

        :param dc: Tushare 数据缓存对象
        :param strategy: 股票择时策略
        :param init_n: 初始化 Trader 需要的最少基础K线数量
        :param sdt: 开始回测时间
        :param edt: 结束回测时间
        """
        self.name = self.__class__.__name__
        self.strategy = strategy
        self.init_n = init_n
        self.data_path = dc.data_path
        self.res_path = os.path.join(self.data_path, strategy.__name__)
        os.makedirs(self.res_path, exist_ok=True)

        file_strategy = os.path.join(self.res_path, f'{strategy.__name__}_strategy.txt')
        with open(file_strategy, 'w', encoding='utf-8') as f:
            f.write(inspect.getsource(strategy))
        print(f"strategy saved into {file_strategy}")

        self.dc = dc
        self.sdt = sdt
        self.edt = edt

        stocks = self.dc.stock_basic()
        stocks_ = stocks[stocks['list_date'] < '2010-01-01'].ts_code.to_list()
        self.stocks_map = {
            "index": ['000905.SH', '000016.SH', '000300.SH', '000001.SH', '000852.SH',
                      '399001.SZ', '399006.SZ', '399376.SZ', '399377.SZ', '399317.SZ', '399303.SZ'],
            "stock": stocks.ts_code.to_list(),
            "check": ['000001.SZ'],
            "train": stocks_[:200],
            "valid": stocks_[200:600],
            "etfs": ['512880.SH', '518880.SH', '515880.SH', '513050.SH', '512690.SH',
                     '512660.SH', '512400.SH', '512010.SH', '512000.SH', '510900.SH',
                     '510300.SH', '510500.SH', '510050.SH', '159992.SZ', '159985.SZ',
                     '159981.SZ', '159949.SZ', '159915.SZ'],
        }

    def analyze_results(self, step, trade_dir="long"):
        res_path = self.res_path
        raw_path = os.path.join(res_path, f'raw_{step}')
        df_ops, df_pairs, df_p = read_raw_results(raw_path, trade_dir)
        s_name = self.strategy.__name__

        df_ops.to_csv(os.path.join(res_path, f"{s_name}_{step}_{trade_dir}_operates.csv"), index=False)
        df_pairs.to_excel(os.path.join(res_path, f"{s_name}_{step}_{trade_dir}_pairs.xlsx"),
                          index=False)

        f = pd.ExcelWriter(os.path.join(res_path, f"{s_name}_{step}_{trade_dir}_performance.xlsx"))
        df_p.to_excel(f, sheet_name="评估", index=False)
        tp = TraderPerformance(df_pairs)
        for col in tp.agg_columns:
            df_ = tp.agg_statistics(col)
            df_.to_excel(f, sheet_name=f"{col}聚合", index=False)
        f.close()
        print(f"{s_name} - {step} - {trade_dir}: {tp.basic_info}")

    def batch_backtest(self, step):
        """批量回测

        :param step: 择时策略研究阶段
            check   在给定的股票上观察策略交易的准确性，输出交易快照
            index   在股票指数上评估策略表现
            train   在训练集上评估策略表现
            valid   在验证集上评估策略表现
            stock   用全市场所有股票评估策略表现
        :return:
        """
        assert step in self.stocks_map.keys(), f"step 参数输入错误，可选值：{list(self.stocks_map.keys())}"

        init_n = self.init_n
        save_html = True if step == 'check' else False
        ts_codes = self.stocks_map[step]
        dc = self.dc
        res_path = self.res_path
        sdt = self.sdt
        edt = self.edt
        strategy = self.strategy
        raw_path = os.path.join(res_path, f"raw_{step}")
        os.makedirs(raw_path, exist_ok=True)

        if step == 'index':
            asset = "I"
        else:
            asset = 'E'

        tactic = strategy()
        base_freq = tactic['base_freq']
        signals_n = tactic.get('signals_n', 0)
        assert signals_n >= 0

        with open(os.path.join(res_path, f'{strategy.__name__}_strategy.txt'), 'w', encoding='utf-8') as f:
            f.write(inspect.getsource(strategy))

        for ts_code in ts_codes:
            if save_html:
                html_path = os.path.join(res_path, f"raw_{step}/{ts_code}")
                os.makedirs(html_path, exist_ok=True)
            else:
                html_path = None

            try:
                file_res = os.path.join(res_path, f"raw_{step}/{ts_code}.xlsx")
                file_signals = os.path.join(res_path, f"raw_{step}/{ts_code}_signals.pkl")
                if os.path.exists(file_res) and os.path.exists(file_signals):
                    print(f"exits: {file_res}")
                    continue

                if "分钟" in base_freq:
                    bars = dc.pro_bar_minutes(ts_code, sdt, edt, freq=freq_cn2ts[base_freq],
                                              asset=asset, adj='hfq', raw_bar=True)
                else:
                    bars = dc.pro_bar(ts_code, sdt, edt, freq=freq_cn2ts[base_freq],
                                      asset=asset, adj='hfq', raw_bar=True)
                res = trader_fast_backtest(bars, init_n, strategy, html_path, signals_n=signals_n)

                # 保存信号结果
                dfs = pd.DataFrame(res['signals'])
                c_cols = [k for k, v in dfs.dtypes.to_dict().items() if v.name.startswith('object')]
                dfs[c_cols] = dfs[c_cols].astype('category')
                float_cols = [k for k, v in dfs.dtypes.to_dict().items() if v.name.startswith('float')]
                dfs[float_cols] = dfs[float_cols].astype('float32')
                dfs.to_pickle(file_signals, protocol=4)

                f = pd.ExcelWriter(file_res)
                if res.get('long_performance', None):
                    print(f"{strategy.__name__} long_performance: {res['long_performance']}")
                    pd.DataFrame(res['long_holds']).to_excel(f, sheet_name="long_holds", index=False)
                    pd.DataFrame(res['long_operates']).to_excel(f, sheet_name="long_operates", index=False)
                    pd.DataFrame(res['long_pairs']).to_excel(f, sheet_name="long_pairs", index=False)
                    pd.DataFrame([res['long_performance']]).to_excel(f, sheet_name="long_performance", index=False)

                if res.get('short_performance', None):
                    print(f"{strategy.__name__} short_performance: {res['short_performance']}")
                    pd.DataFrame(res['short_holds']).to_excel(f, sheet_name="short_holds", index=False)
                    pd.DataFrame(res['short_operates']).to_excel(f, sheet_name="short_operates", index=False)
                    pd.DataFrame(res['short_pairs']).to_excel(f, sheet_name="short_pairs", index=False)
                    pd.DataFrame([res['short_performance']]).to_excel(f, sheet_name="short_performance", index=False)

                f.close()
            except:
                traceback.print_exc()

        if tactic.get('long_events', None):
            self.analyze_results(step, 'long')
        if tactic.get('short_events', None):
            self.analyze_results(step, 'short')
        print(f"results saved into {self.res_path}")

    def analyze_signals(self, step: str):
        """分析策略中信号的基础表现

        :param step:
        :return:
        """
        dc = self.dc
        raw_path = os.path.join(self.res_path, f"raw_{step}")
        file_dfs = os.path.join(self.res_path, f"{step}_dfs.pkl")
        signals_pat = fr"{raw_path}\*_signals.pkl"
        freq = freq_cn2ts[self.strategy()['base_freq']]

        # 由于python存在循环导入的问题，只能把两个导入放到这里
        from ..sensors.utils import read_cached_signals, SignalsPerformance

        if not os.path.exists(file_dfs):
            dfs = read_cached_signals(file_dfs, signals_pat)
            asset = "I" if step == 'index' else "E"
            results = []
            for symbol, dfg in tqdm(dfs.groupby('symbol'), desc='add nbar'):
                dfk = dc.pro_bar_minutes(symbol, sdt=dfg['dt'].min(), edt=dfg['dt'].max(),
                                         freq=freq, asset=asset, adj='hfq', raw_bar=False)
                dfk_cols = ['dt'] + [x for x in dfk.columns if x not in dfs.columns]
                dfk = dfk[dfk_cols]
                dfs_ = dfg.merge(dfk, on='dt', how='left')
                results.append(dfs_)

            dfs = pd.concat(results, ignore_index=True)
            c_cols = [k for k, v in dfs.dtypes.to_dict().items() if v.name.startswith('object')]
            dfs[c_cols] = dfs[c_cols].astype('category')
            float_cols = [k for k, v in dfs.dtypes.to_dict().items() if v.name.startswith('float')]
            dfs[float_cols] = dfs[float_cols].astype('float32')
            dfs.to_pickle(file_dfs, protocol=4)
        else:
            dfs = pd.read_pickle(file_dfs)

        results_path = os.path.join(raw_path, 'signals_performance')
        os.makedirs(results_path, exist_ok=True)
        signal_cols = [x for x in dfs.columns if len(x.split("_")) == 3]
        for key in signal_cols:
            file_xlsx = os.path.join(results_path, f"{key.replace(':', '')}.xlsx")
            sp = SignalsPerformance(dfs, keys=[key], dc=dc)
            sp.report(file_xlsx)
            print(f"{key} performance saved into {file_xlsx}")
